// PV112 2017, lesson 6 - uniform buffer objects, instancing

#include <iostream>
#include "PV112.h"

#define _USE_MATH_DEFINES
#include <math.h>

// Include GLEW, use static library
#define GLEW_STATIC
#include <GL/glew.h>
#pragma comment(lib, "glew32s.lib") // Link with GLEW library
#pragma comment(lib, "DevIL.lib") // Link with DevIL library

// Include FreeGLUT
#include <GL/freeglut.h>

// Include the most important GLM functions
#include <glm/gtc/type_ptr.hpp>
#include <glm/gtc/matrix_transform.hpp>

using namespace std;
using namespace PV112;

// Current window size
int win_width = 480;
int win_height = 480;

// Shader program for the cube and its uniforms
GLuint cube_program;

GLint cube_model_matrix_loc;
GLint cube_view_matrix_loc;
GLint cube_projection_matrix_loc;

GLint cube_material_specular_color_loc;
GLint cube_material_shininess_loc;

GLint cube_light_position_loc;
GLint cube_light_ambient_color_loc;
GLint cube_light_diffuse_color_loc;
GLint cube_light_specular_color_loc;

GLint cube_eye_position_loc;

GLint cube_wood_tex_loc;

// Shader program for the teapot and its uniforms
GLuint teapot_program;

GLint teapot_model_matrix_loc;
GLint teapot_view_matrix_loc;
GLint teapot_projection_matrix_loc;

GLint teapot_material_color_loc;
GLint teapot_material_specular_color_loc;
GLint teapot_material_shininess_loc;

GLint teapot_light_position_loc;
GLint teapot_light_ambient_color_loc;
GLint teapot_light_diffuse_color_loc;
GLint teapot_light_specular_color_loc;

GLint teapot_eye_position_loc;

// Simple geometries that we will use in this lecture
PV112Geometry my_cube;
PV112Geometry my_sphere;
PV112Geometry my_teapot;

// Simple camera that allows us to look at the object from different views
PV112Camera my_camera;

// Wood texture ID
GLuint wood_tex;

// Data of our light
struct Light
{
    glm::vec4 position;
    glm::vec4 ambient_color;
    glm::vec4 diffuse_color;
    glm::vec4 specular_color;
};
Light LightData; // Data of the light on the CPU
GLuint LightData_ubo; // Data of the light on GPU in a uniform buffer object

// Data of our camera
struct Camera
{
    glm::mat4 view_matrix;
    glm::mat4 projection_matrix;
    glm::vec3 eye_position;
};
Camera CameraData;     // Data of the camera on the CPU
GLuint CameraData_ubo; // Data of the camera on the GPU in a uniform buffer object

// Positions of all instanced cubes
static const int INSTANCES_COUNT = 100;
struct Cubes
{
    glm::mat4 model_matrix[INSTANCES_COUNT];
};
Cubes CubeData; // Data of the cubes on CPU
GLuint CubeData_ubo; // Data of the cubes on GPU in a uniform buffer object

// Positions and colors of all instanced teapots
struct Teapots
{
    glm::mat4 model_matrix[INSTANCES_COUNT];
    glm::vec4 color[INSTANCES_COUNT];
};
Teapots TeapotData; // Data of the teapots on CPU
GLuint TeapotData_ubo; // Data of the teapots on GPU in a uniform buffer object

// Current time of the application in seconds, for animations
float app_time_s = 0.0f;

// Called when the user presses a key
void key_pressed(unsigned char key, int mouseX, int mouseY)
{
    switch (key)
    {
    case 27: // Escape
        exit(0);
        break;
    case 'l':
        glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
        glutPostRedisplay();
        break;
    case 'f':
        glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
        glutPostRedisplay();
        break;
    case 't':
        glutFullScreenToggle();
        break;
    }
}

// Called when the user presses a mouse button
void mouse_button_changed(int button, int state, int x, int y)
{
    my_camera.OnMouseButtonChanged(button, state, x, y);
}

// Called when the user moves with the mouse (when some mouse button is pressed)
void mouse_moved(int x, int y)
{
    my_camera.OnMouseMoved(x, y);
}

// Draws PV112Geometry like DrawGeometry, it only uses instaned drawing
void DrawGeometryInstanced(const PV112Geometry &geom, int primcount)
{
    if (geom.DrawArraysCount > 0)
        glDrawArraysInstanced(geom.Mode, 0, geom.DrawArraysCount, primcount);
    if (geom.DrawElementsCount > 0)
        glDrawElementsInstanced(geom.Mode, geom.DrawElementsCount,
                GL_UNSIGNED_INT, nullptr, primcount);
}

// Initializes OpenGL stuff
void init()
{
    glClearColor(0.0f, 0.0f, 0.0f, 0.0f);
    glClearDepth(1.0);
    glEnable(GL_DEPTH_TEST);

    // Force indices of input variables
    int position_loc = 0;
    int normal_loc = 1;
    int tex_coord_loc = 2;

    //------------
    //--  Cube  --

    // Create shader program
    cube_program = CreateAndLinkProgram("cube_vertex.glsl", "cube_fragment.glsl",
        position_loc, "position", normal_loc, "normal", tex_coord_loc, "tex_coord");
    if (0 == cube_program)
        WaitForEnterAndExit();

    // Get variable locations
    cube_model_matrix_loc = glGetUniformLocation(cube_program, "model_matrix");
    cube_view_matrix_loc = glGetUniformLocation(cube_program, "view_matrix");
    cube_projection_matrix_loc = glGetUniformLocation(cube_program, "projection_matrix");
    cube_material_specular_color_loc = glGetUniformLocation(cube_program, "material_specular_color");
    cube_material_shininess_loc = glGetUniformLocation(cube_program, "material_shininess");
    cube_light_position_loc = glGetUniformLocation(cube_program, "light_position");
    cube_light_diffuse_color_loc = glGetUniformLocation(cube_program, "light_diffuse_color");
    cube_light_ambient_color_loc = glGetUniformLocation(cube_program, "light_ambient_color");
    cube_light_specular_color_loc = glGetUniformLocation(cube_program, "light_specular_color");
    cube_eye_position_loc = glGetUniformLocation(cube_program, "eye_position");
    cube_wood_tex_loc = glGetUniformLocation(cube_program, "wood_tex");

    // Set the sampler to take the data for the 0-th texture unit
    glUseProgram(cube_program);
    glUniform1i(cube_wood_tex_loc, 0);
    glUseProgram(0);

    //--------------
    //--  Teapot  --

    // Create shader program
    teapot_program = CreateAndLinkProgram("teapot_vertex.glsl", "teapot_fragment.glsl",
        position_loc, "position", normal_loc, "normal", tex_coord_loc, "tex_coord");
    if (0 == teapot_program)
        WaitForEnterAndExit();

    // Get variable locations
    teapot_model_matrix_loc = glGetUniformLocation(teapot_program, "model_matrix");
    teapot_view_matrix_loc = glGetUniformLocation(teapot_program, "view_matrix");
    teapot_projection_matrix_loc = glGetUniformLocation(teapot_program, "projection_matrix");
    teapot_material_color_loc = glGetUniformLocation(teapot_program, "material_color");
    teapot_material_specular_color_loc = glGetUniformLocation(teapot_program, "material_specular_color");
    teapot_material_shininess_loc = glGetUniformLocation(teapot_program, "material_shininess");
    teapot_light_position_loc = glGetUniformLocation(teapot_program, "light_position");
    teapot_light_ambient_color_loc = glGetUniformLocation(teapot_program, "light_ambient_color");
    teapot_light_diffuse_color_loc = glGetUniformLocation(teapot_program, "light_diffuse_color");
    teapot_light_specular_color_loc = glGetUniformLocation(teapot_program, "light_specular_color");
    teapot_eye_position_loc = glGetUniformLocation(teapot_program, "eye_position");

    //-------------
    //--  Light  --

    LightData.position = glm::vec4(0.0f, 0.0f, 0.0f, 1.0f);
    LightData.ambient_color = glm::vec4(0.3f, 0.3f, 0.3f, 1.0f);
    LightData.diffuse_color = glm::vec4(1.0f, 1.0f, 1.0f, 1.0f);
    LightData.specular_color = glm::vec4(1.0f, 1.0f, 1.0f, 1.0f);

    /*
    glGenBuffers(1, &LightData_ubo);
    glBindBuffer(GL_UNIFORM_BUFFER, LightData_ubo);
    glBufferData(GL_UNIFORM_BUFFER, sizeof(Light), &LightData, GL_STATIC_DRAW);
    glBindBuffer(GL_UNIFORM_BUFFER, 0);
    */

    //--------------
    //--  Camera  --

    CameraData.view_matrix = glm::mat4(1.0f);
    CameraData.projection_matrix = glm::mat4(1.0f);
    CameraData.eye_position = glm::vec3(0.0f);

    /*
    glGenBuffers(1, &CameraData_ubo);
    glBindBuffer(GL_UNIFORM_BUFFER, CameraData_ubo);
    glBufferData(GL_UNIFORM_BUFFER, sizeof(Camera), &CameraData, GL_STATIC_DRAW);
    glBindBuffer(GL_UNIFORM_BUFFER, 0);
    */

    //-----------------------
    //--  Instancing data  --
    /*
    srand(12345); // Any seed we want

    // Create the data of all instances
    for (int i = 0; i < INSTANCES_COUNT; i++)
    {
        // Compute the positions of all cubes - create a 10x10 grid
        CubeData.model_matrix[i] = glm::translate(glm::mat4(1.0f),
                glm::vec3(float(i%10 - 4) * 5.0f, 0.0f, float(i/10 - 4) * 5.0f));

        // Compute the position of all teapots - place them on the cubes and randomly rotate
        //TeapotData.model_matrix[i] = CubeData.model_matrix[i] *
        //    glm::translate(glm::mat4(1.0f), glm::vec3(0.0f, 1.0f, 0.0f)) *
        //    glm::rotate(glm::mat4(1.0f), float(rand()) / float(RAND_MAX) * 2.0f * float(M_PI),
        //            glm::vec3(0.0f, 1.0f, 0.0f));

        // Choose a random color for each cube
        //TeapotData.color[i] = glm::vec4(glm::vec3(float(rand()),
        //    float(rand()), float(rand())) / float(RAND_MAX), 1.0f);
    }

    glGenBuffers(1, &CubeData_ubo);
    glBindBuffer(GL_UNIFORM_BUFFER, CubeData_ubo);
    glBufferData(GL_UNIFORM_BUFFER, sizeof(Cubes), &CubeData, GL_STATIC_DRAW);
    //glGenBuffers(1, &TeapotData_ubo);
    //glBindBuffer(GL_UNIFORM_BUFFER, TeapotData_ubo);
    //glBufferData(GL_UNIFORM_BUFFER, sizeof(Teapots), &TeapotData, GL_STATIC_DRAW);
    glBindBuffer(GL_UNIFORM_BUFFER, 0);
    */
    //-------------
    //--  Other  --

    // Create geometries
    my_cube = CreateCube(position_loc, normal_loc, tex_coord_loc);
    my_sphere = CreateSphere(position_loc, normal_loc, tex_coord_loc);
    my_teapot = CreateTeapot(position_loc, normal_loc, tex_coord_loc);

    // Load and set the parameters of a wood texture
    wood_tex = CreateAndLoadTexture2D(MAYBEWIDE("wood.jpg"));
    glBindTexture(GL_TEXTURE_2D, wood_tex);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    glBindTexture(GL_TEXTURE_2D, 0);
}

// Called when the window needs to be rerendered
void render()
{
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

    // Some variables
    glm::mat4 model_matrix;

    //-------------
    //--  Light  --

    // Light position, with a simple animation
    LightData.position =
        glm::rotate(glm::mat4(1.0f), app_time_s, glm::vec3(0.0f, 1.0f, 0.0f)) *
        glm::translate(glm::mat4(1.0f), glm::vec3(30.0f, 10.0f, 0.0f)) *
        glm::vec4(0.0f, 0.0f, 0.0f, 1.0f);

    /*
    glBindBuffer(GL_UNIFORM_BUFFER, LightData_ubo);
    glBufferSubData(GL_UNIFORM_BUFFER, 0, sizeof(Light), &LightData);
    glBindBuffer(GL_UNIFORM_BUFFER, 0);
    */

    //--------------
    //--  Camera  --

    // Camera matrices and eye position
    CameraData.projection_matrix = glm::perspective(glm::radians(45.0f),
            float(win_width) / float(win_height), 0.1f, 1000.0f);
    CameraData.view_matrix = glm::lookAt(my_camera.GetEyePosition(),
            glm::vec3(0.0f, 0.0f, 0.0f), glm::vec3(0.0f, 1.0f, 0.0f));
    CameraData.eye_position = my_camera.GetEyePosition();

    /*
    glBindBuffer(GL_UNIFORM_BUFFER, CameraData_ubo);
    glBufferSubData(GL_UNIFORM_BUFFER, 0, sizeof(Camera), &CameraData);
    glBindBuffer(GL_UNIFORM_BUFFER, 0);
    */

    //-------------
    //--  Cubes  --

    // Use the cube shader program
    glUseProgram(cube_program);

    // Set camera data (projection matrix, view matrix and position of the eye)
    glUniformMatrix4fv(cube_projection_matrix_loc, 1, GL_FALSE, glm::value_ptr(CameraData.projection_matrix));
    glUniformMatrix4fv(cube_view_matrix_loc, 1, GL_FALSE, glm::value_ptr(CameraData.view_matrix));
    glUniform3fv(cube_eye_position_loc, 1, glm::value_ptr(CameraData.eye_position));

    // Set light data (position and colors)
    glUniform4fv(cube_light_position_loc,       1, glm::value_ptr(LightData.position));
    glUniform4fv(cube_light_ambient_color_loc,  1, glm::value_ptr(LightData.ambient_color));
    glUniform4fv(cube_light_diffuse_color_loc,  1, glm::value_ptr(LightData.diffuse_color));
    glUniform4fv(cube_light_specular_color_loc, 1, glm::value_ptr(LightData.specular_color));

    // Set model's material and texture
    glUniform3f(cube_material_specular_color_loc, 1.0f, 1.0f, 1.0f);
    glUniform1f(cube_material_shininess_loc, 40.0f);
    glActiveTexture(GL_TEXTURE0);
    glBindTexture(GL_TEXTURE_2D, wood_tex);

    // Set model's model matrix
    model_matrix = glm::mat4(1.0f);
    glUniformMatrix4fv(cube_model_matrix_loc, 1, GL_FALSE, glm::value_ptr(model_matrix));

    // Draw the cube
    glBindVertexArray(my_cube.VAO);
    DrawGeometry(my_cube);

    //---------------
    //--  Teapots  --

    // Use the teapot shader program
    glUseProgram(teapot_program);

    // Set camera data (projection matrix, view matrix and position of the eye)
    glUniformMatrix4fv(teapot_projection_matrix_loc, 1, GL_FALSE, glm::value_ptr(CameraData.projection_matrix));
    glUniformMatrix4fv(teapot_view_matrix_loc, 1, GL_FALSE, glm::value_ptr(CameraData.view_matrix));
    glUniform3fv(teapot_eye_position_loc, 1, glm::value_ptr(CameraData.eye_position));

    // Set light data (position and colors)
    glUniform4fv(teapot_light_position_loc,       1, glm::value_ptr(LightData.position));
    glUniform4fv(teapot_light_ambient_color_loc,  1, glm::value_ptr(LightData.ambient_color));
    glUniform4fv(teapot_light_diffuse_color_loc,  1, glm::value_ptr(LightData.diffuse_color));
    glUniform4fv(teapot_light_specular_color_loc, 1, glm::value_ptr(LightData.specular_color));

    // Set model's color
    glUniform4f(teapot_material_color_loc, 0.0f, 0.0f, 1.0f, 1.0f);
    glUniform3f(teapot_material_specular_color_loc, 1.0f, 1.0f, 1.0f);
    glUniform1f(teapot_material_shininess_loc, 40.0f);

    // Set model's model matrix
    model_matrix = glm::translate(glm::mat4(1.0f), glm::vec3(0.0f, 1.0f, 0.0f));
    glUniformMatrix4fv(teapot_model_matrix_loc, 1, GL_FALSE, glm::value_ptr(model_matrix));

    // Draw the teapot
    glBindVertexArray(my_teapot.VAO);
    DrawGeometry(my_teapot);

    //-------------------------------------
    //--  Reset the VAO and the program  --

    glBindVertexArray(0);
    glUseProgram(0);

    // Swaps the front and back buffer (double-buffering)
    glutSwapBuffers();
}

// Called when the window changes its size
void reshape(int width, int height)
{
    win_width = width;
    win_height = height;

    glViewport(0, 0, win_width, win_height);
}

// Callback function to be called when we make an error in OpenGL
void GLAPIENTRY simple_debug_callback(GLenum source, GLenum type, GLuint id,
        GLenum severity, GLsizei length, const char* message, const void* userParam)
{
    switch (type) // Report only some errors
    {
    case GL_DEBUG_TYPE_ERROR:
    case GL_DEBUG_TYPE_DEPRECATED_BEHAVIOR:
    case GL_DEBUG_TYPE_UNDEFINED_BEHAVIOR:
        cout << message << endl; // Put the breakpoint here
        return;
    default:
        return;
    }
}

// Simple timer function for animations
void timer(int)
{
    // Add 20 ms to the application time, register the timer to be called again
    // in 20 ms, and rerender the scene
    app_time_s += 0.020f;
    glutTimerFunc(20, timer, 0);
    glutPostRedisplay();
}

int main(int argc, char ** argv)
{
    // Initialize GLUT
    glutInit(&argc, argv);
    glutInitDisplayMode(GLUT_DEPTH | GLUT_DOUBLE | GLUT_RGBA);
    glutSetOption(GLUT_ACTION_ON_WINDOW_CLOSE, GLUT_ACTION_GLUTMAINLOOP_RETURNS);

    // Set OpenGL Context parameters
    glutInitContextVersion(3, 3);              // Use OpenGL 3.3
    glutInitContextProfile(GLUT_CORE_PROFILE); // Use OpenGL core profile
    glutInitContextFlags(GLUT_DEBUG);          // Use OpenGL debug context

    // Initialize window
    glutInitWindowSize(win_width, win_height);
    glutCreateWindow("PV112 - cv6");

    // Initialize GLEW
    glewExperimental = GL_TRUE;
    glewInit();

    // Initialize image library
    InitImageLibrary();

    // Set the debug callback
    SetDebugCallback(simple_debug_callback);

    // Initialize out OpenGL stuff
    init();

    // Register callbacks
    glutDisplayFunc(render);
    glutReshapeFunc(reshape);
    glutKeyboardFunc(key_pressed);
    glutTimerFunc(20, timer, 0);
    glutMouseFunc(mouse_button_changed);
    glutMotionFunc(mouse_moved);

    // Run the main loop
    glutMainLoop();

    return 0;
}
